/*
 * Internal bit-depth increase (IBDI) and memory compression
 *
 * IBDI off: 8 bit to 4 bit compression/decompression of reference picture
 * IBDI on: 12 bit to 8 bit compression/decompression of reference picture
 *
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 * ALL RIGHTS RESERVED 
 *                                                                                                                                                                                                                                                                     
*/


#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "global.h"
#include "mbuffer.h"
#include "image.h"

typedef struct CFB_bitstream {
  int bit_ptr;
  int buf_ptr;
  int bufSize;
  unsigned char *databuf;
}CFB_Bitstream;

static unsigned int bhMask[17] = { 0x0000,
  0x0001, 0x0003, 0x0007, 0x000f,
  0x001f, 0x003f, 0x007f, 0x00ff,
  0x01ff, 0x03ff, 0x07ff, 0x0fff,
  0x1fff, 0x3fff, 0x7fff, 0xffff
};

void CompressDecompressFrame(StorablePicture *);
void CompressFrame(StorablePicture *, CFB_Bitstream *);
void DecompressFrame(StorablePicture *, CFB_Bitstream *);
void LoadBlock(int, int, int, unsigned short **, unsigned short **, unsigned short **, short **, short **, short **);
void StoreBlock(int, int, int, unsigned short **, unsigned short **, unsigned short **, short **, short **, short **);
void Forward2DTransform(int, int, short **);
void UnrolledForward2DTransform(int, int, short **);
void Backward2DTransform(int, int, short **);
void UnrolledBackward2DTransform(int, int, short **);
void Quantization(int, int, short **, short **);
void InverseQuantization(int, int, short **, short **);
int ComputeEntropyCodedBitSizeY(short **, int *);
int ComputeEntropyCodedBitSizeUV(short **, short **, int *);
void ComputeDCCoefficientResidual(int, int, short**, short**, short**, short**, short**, short**);
void ReconstructDCCoefficient(int, int, short**, short**, short**, short**, short**, short**);
void WriteCompressedBlock(CFB_Bitstream *, short**, short**, short**, int*, int*, int*);
void ReadCompressedBlock(CFB_Bitstream *, short**, short**, short**, int*, int*, int*);
void UnsignedExpGolombEncoding(CFB_Bitstream *, int, unsigned int);
unsigned int UnsignedExpGolombDecoding(CFB_Bitstream * stream, int);
void PutBits(CFB_Bitstream*, int, unsigned int);
unsigned int ReadBits(CFB_Bitstream*, int);
unsigned int NextBits(CFB_Bitstream*, int);
void FlushBits(CFB_Bitstream* , int);
unsigned int ilog2(register unsigned int);


#define minimumImageSizeForCFB 0 //1280*720 // mimumum image size to enable coded picture compression (720p)
#define blockHeight 8 // NxN pixels: size of a block (luma), should be 8
//#define minCompressionRatio 2 // minimum compression ratio
#define MBHeight 64  // NxN pixels: MB size
#define numberOfQuantizationIndices 16 // number of quantization indices
#define initialRemainingBitsBufferParameter 4 // parameter for setting initial fullness of the remaining bits buffer
#define bypassEntropyCoding 1 // parameter for bypassing entropy coding. Used for compliant decoder that does not do memory compression.

/*!
************************************************************************
* \brief
*    Compresses the input YUV420 image and writes decompressed output
*    YUV420 image. Compression ratio is fixed for a block or group 
*    of blocks.
*
*  \par Input:
*     p: encoded picture in StorablePicture format
*
************************************************************************
*/
void CompressDecompressFrame(StorablePicture *p)
{
  int imageWidth;
  int imageHeight;
  double minCompressionRatio; // minimum compression ratio
  CFB_Bitstream compressedFrameBuf;
  

  // Use 12 bit to 8 bit compression for IBDI on
  if (img->bitdepth_luma >= 12)
  {
    minCompressionRatio = 1.5; 
  }
  else // Use 8 bit to 4 bit compression for IBDI off
  {
    minCompressionRatio = 2; 
  }

  imageWidth = p->size_x;
  imageHeight = p->size_y;
  // Coded frame compression is disabled for image sizes less than minimumImageSizeForCFB
  if(imageWidth*imageHeight < minimumImageSizeForCFB)
  {
    return;
  }

  // Initialize compressed frame buffer
  compressedFrameBuf.bit_ptr = 8;
  compressedFrameBuf.buf_ptr = 0;
  compressedFrameBuf.bufSize = (int) ((imageWidth*imageHeight*(img->bitdepth_luma + img->bitdepth_chroma/2))/(8*minCompressionRatio));
#if bypassEntropyCoding == 0
  compressedFrameBuf.databuf = (unsigned char *)calloc(compressedFrameBuf.bufSize,1);
#endif

  CompressFrame(p, &compressedFrameBuf);

// Reconstruction is performed withing CompressFrame function when bypassEntropyCoding flag is set 
#if bypassEntropyCoding == 0
  DecompressFrame(p, &compressedFrameBuf);
  // free memory for compressed frame buffer
  free(compressedFrameBuf.databuf);
#endif
  return;
}

/*!
************************************************************************
* \brief
*    Compresses the input YUV420 image
*    Compression ratio is fixed for a block or group of blocks.
*
*  \par Input:
*     p: encoded picture in StorablePicture format
*  \par Output:
*     compressedFrameBuf: pointer compressed frame buffer
*
************************************************************************
*/

void CompressFrame(StorablePicture *p,
                   CFB_Bitstream *compressedFrameBuf)
{
  int tempbuf, tempbit;
  int i, j;
  int imageWidth;
  int imageHeight;
  unsigned short ** inputFrameY;
  unsigned short ** inputFrameU;
  unsigned short ** inputFrameV;
  int numberOfBlockRows, numberOfBlockColumns;
  int MBHeightInBlocks;
  int numberOfTransformLevelsLuma, numberOfTransformLevelsChroma; 
  int blockBitBudget, remainingBlockBitBudget, totalBlockBitSize, lumaBitSize, chromaBitSize,quantizationLevelBitSize, EGCodingModeFlagsSize;

  short **blockY, **blockYQ;
  short **blockU, **blockUQ;
  short **blockV, **blockVQ;

  short **YDCCoeffBuffer;
  short **UDCCoeffBuffer;
  short **VDCCoeffBuffer;

  int QIdx;
  int selectedQIdx, EGKLuma, EGKChroma;
  int compressionRateGroup; // 0: fixed compression rate for each block 1: fixed compression rate for a row of MBs
  double minCompressionRatio; // minimum compression ratio
  // Use block based compression for IBDI on
  // Use 12 bit to 8 bit compression for IBDI on
  if (img->bitdepth_luma >= 12)
  {
    compressionRateGroup = 0;
    minCompressionRatio = 1.5; 
  }
  else // use MB row compression for IBDI off
       // Use 8 bit to 4 bit compression for IBDI off
  {
    compressionRateGroup = 1;
    minCompressionRatio = 2; 
  }

  imageWidth = p->size_x;
  imageHeight = p->size_y;
  inputFrameY = p->imgY;
  inputFrameU = p->imgUV[0];
  inputFrameV = p->imgUV[1];

  numberOfBlockRows = imageHeight/blockHeight;
  numberOfBlockColumns = imageWidth/blockHeight;
  MBHeightInBlocks = (MBHeight/blockHeight);

  // Compute blockBitBudget for luma+chroma using blockHeight and minimumCompression ratio
  blockBitBudget = (int) ((blockHeight*blockHeight*(img->bitdepth_luma + img->bitdepth_chroma/2))/minCompressionRatio);
  remainingBlockBitBudget = blockBitBudget;

  numberOfTransformLevelsLuma = ilog2(blockHeight);
  numberOfTransformLevelsChroma = numberOfTransformLevelsLuma - 1;


  // allocate memory for blocks
  // Y
  blockY = (short **) malloc(blockHeight*sizeof(short *));
  blockYQ = (short **) malloc(blockHeight*sizeof(short *));
  for(i=0; i<blockHeight; i++)
  {
    blockY[i] = (short *) malloc(blockHeight*sizeof(short));
    blockYQ[i] = (short *) malloc(blockHeight*sizeof(short));
  }
  // U and V
  blockU = (short **) malloc((blockHeight/2)*sizeof(short *));
  blockUQ = (short **) malloc((blockHeight/2)*sizeof(short *));
  blockV = (short **) malloc((blockHeight/2)*sizeof(short *));
  blockVQ = (short **) malloc((blockHeight/2)*sizeof(short *));
  for(i=0; i<blockHeight/2; i++)
  {
    blockU[i] = (short *) malloc((blockHeight/2)*sizeof(short));
    blockUQ[i] = (short *) malloc((blockHeight/2)*sizeof(short));
    blockV[i] = (short *) malloc((blockHeight/2)*sizeof(short));
    blockVQ[i] = (short *) malloc((blockHeight/2)*sizeof(short));
  }

  // Allocate memory for DC coefficient buffers
  YDCCoeffBuffer = (short **) malloc(numberOfBlockRows*sizeof(short *));
  UDCCoeffBuffer = (short **) malloc(numberOfBlockRows*sizeof(short *));
  VDCCoeffBuffer = (short **) malloc(numberOfBlockRows*sizeof(short *));
  for(i=0; i<numberOfBlockRows; i++)
  {
    YDCCoeffBuffer[i] = (short *) malloc(numberOfBlockColumns*sizeof(short));
    UDCCoeffBuffer[i] = (short *) malloc(numberOfBlockColumns*sizeof(short));
    VDCCoeffBuffer[i] = (short *) malloc(numberOfBlockColumns*sizeof(short));
  }

  // loop over block rows
  for(i=0; i<numberOfBlockRows; i++)
  {
    // for the fixed MB row option
    if(compressionRateGroup)
    {
      // initialize the remainingBlockBitBudget at the MB row start
      if( i % MBHeightInBlocks == 0)
      {
        // Seek the bistream pointers
        compressedFrameBuf->bit_ptr = 8;
        compressedFrameBuf->buf_ptr = ( (blockBitBudget*i*numberOfBlockColumns)>>3 );
        
        // calculate the MBHeightInBlocks for the last MB row
        if( i == MBHeightInBlocks*((int) (numberOfBlockRows/MBHeightInBlocks)) )
        {
          MBHeightInBlocks = numberOfBlockRows - (MBHeightInBlocks)*((int) (numberOfBlockRows/MBHeightInBlocks));
        }

        // initialize remaining bit budget
        remainingBlockBitBudget = blockBitBudget + initialRemainingBitsBufferParameter*(numberOfBlockColumns*MBHeightInBlocks-1);      
      }
    }

    // loop over block columns
    for(j=0; j<numberOfBlockColumns; j++)
    {
      // Seek the bistream pointers for the fixed block size option
      if(!compressionRateGroup)
      {
        // Seek the bistream pointers
        compressedFrameBuf->bit_ptr = 8;
        compressedFrameBuf->buf_ptr = ( (blockBitBudget*(i*numberOfBlockColumns+j))>>3 );
      }

      // load the block data form the input subtract 128
      LoadBlock(i, j, imageWidth, inputFrameY, inputFrameU, inputFrameV, blockY, blockU, blockV);

      // perform 2-D DWT
      // Y
      UnrolledForward2DTransform( blockHeight, numberOfTransformLevelsLuma, blockY);
      // U
      UnrolledForward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockU);
      // V
      UnrolledForward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockV);

      // Store the DC values to buffer 
      YDCCoeffBuffer[i][j] = blockY[0][0];
      UDCCoeffBuffer[i][j] = blockU[0][0];
      VDCCoeffBuffer[i][j] = blockV[0][0];

      if (compressionRateGroup)
      {
        // Compute DC coefficient residual for Y,U,V
        ComputeDCCoefficientResidual(i, j, blockY, blockU, blockV, YDCCoeffBuffer, UDCCoeffBuffer, VDCCoeffBuffer);
      }

      selectedQIdx = numberOfQuantizationIndices -1;

      for(QIdx=0; QIdx<numberOfQuantizationIndices; QIdx++)
      {
        if (QIdx > 0)
        {
          // Quantize the transform coeff
          // Y
          Quantization(QIdx, blockHeight, blockY, blockYQ);
          // U
          Quantization(QIdx, blockHeight/2, blockU, blockUQ);
          // V
          Quantization(QIdx, blockHeight/2, blockV, blockVQ);

          // Compute entropy coded bit size of Y coefficients
          lumaBitSize = ComputeEntropyCodedBitSizeY(blockYQ, &EGKLuma);
          // Compute entropy coded bit size of U and V coefficients
          chromaBitSize = ComputeEntropyCodedBitSizeUV(blockUQ, blockVQ, &EGKChroma);
        }
        else
        {
          // Compute entropy coded bit size of Y coefficients
          lumaBitSize = ComputeEntropyCodedBitSizeY(blockY, &EGKLuma);
          // Compute entropy coded bit size of U and V coefficients
          chromaBitSize = ComputeEntropyCodedBitSizeUV(blockU, blockV, &EGKChroma);
        }

        // Compute quantization level bit size (Unary coding)
        quantizationLevelBitSize = QIdx+1;
        // Exp-Golomb mode selection flag (EG0 or EG3) size 1-bit for Y and 1-bit for UV
        EGCodingModeFlagsSize = 2*1;

        // Compute total block bit size
        totalBlockBitSize = quantizationLevelBitSize + EGCodingModeFlagsSize + lumaBitSize + chromaBitSize;

        if (totalBlockBitSize <= remainingBlockBitBudget)
        {
          // if the block bit size is not larger than the budget, select the QIdx and exit 
          selectedQIdx = QIdx;
          // add unused bits to the remaining bitbudget if compressionRateGroup is set to 1
          if(compressionRateGroup)
          {
            remainingBlockBitBudget += blockBitBudget - totalBlockBitSize - initialRemainingBitsBufferParameter;
          }

          break;
        }
      }

#if bypassEntropyCoding == 0
      tempbuf = compressedFrameBuf->buf_ptr;
      tempbit = compressedFrameBuf->bit_ptr;

      // Entropy encoding
      if ( selectedQIdx > 0 )
      {
        WriteCompressedBlock(compressedFrameBuf, blockYQ, blockUQ, blockVQ, &selectedQIdx, &EGKLuma, &EGKChroma);
      }
      else
      {
        WriteCompressedBlock(compressedFrameBuf, blockY, blockU, blockV, &selectedQIdx, &EGKLuma, &EGKChroma);
      }
#else
      // Do not need to perform inverse operations if no quantization is performed
      if ( selectedQIdx > 0 )
      {
        // Reconstruct DC coefficients for Y,U,V
        blockYQ[0][0] = YDCCoeffBuffer[i][j];
        blockUQ[0][0] = UDCCoeffBuffer[i][j];
        blockVQ[0][0] = VDCCoeffBuffer[i][j];


        // Inverse quantize the transform coeff
        // Y
        InverseQuantization(selectedQIdx, blockHeight, blockY, blockYQ);
        // U
        InverseQuantization(selectedQIdx, blockHeight/2, blockU, blockUQ);
        // V
        InverseQuantization(selectedQIdx, blockHeight/2, blockV, blockVQ);

        // perform inverse 2-D DWT
        // Y
        UnrolledBackward2DTransform( blockHeight, numberOfTransformLevelsLuma, blockY);
        // U
        UnrolledBackward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockU);
        // V
        UnrolledBackward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockV);

        // store the block data to the image array and add 128
        StoreBlock(i, j, imageWidth, inputFrameY, inputFrameU, inputFrameV, blockY, blockU, blockV);
      }
#endif

    }
  }

  // Free memory for blocks
  // Y
  for(i=0; i<blockHeight; i++)
  {
    //free(blockY[i]);
    free(blockYQ[i]);
  }
  free(blockY);
  free(blockYQ);

  // U and V
  for(i=0; i<blockHeight/2; i++)
  {
    //free(blockU[i]);
    free(blockUQ[i]);
    //free(blockV[i]);
    free(blockVQ[i]);
  }
  free(blockU);
  free(blockUQ);
  free(blockV);
  free(blockVQ);


  // free memory for DC coefficient buffers
  for(i=0; i<numberOfBlockRows; i++)
  {
    free(YDCCoeffBuffer[i]);
    free(UDCCoeffBuffer[i]);
    free(VDCCoeffBuffer[i]);
  }
  free(YDCCoeffBuffer);
  free(UDCCoeffBuffer);
  free(VDCCoeffBuffer);

  return;
}

/*!
************************************************************************
* \brief
*    Decompresses the the YUV420 image
*    Compression ratio is fixed for a block or group of blocks.
*
*  
*  \par Input:
*     p: encoded picture in StorablePicture format
*     compressedFrameBuf: pointer compressed frame buffer
*
************************************************************************
*/

void DecompressFrame(StorablePicture *p,
                     CFB_Bitstream *compressedFrameBuf)
{
  int i, j;
  int imageWidth;
  int imageHeight;
  unsigned short ** inputFrameY;
  unsigned short ** inputFrameU;
  unsigned short ** inputFrameV;
  int numberOfBlockRows, numberOfBlockColumns;
  int MBHeightInBlocks;
  int numberOfTransformLevelsLuma, numberOfTransformLevelsChroma; 
  int blockBitBudget;

  short **blockY, **blockYQ;
  short **blockU, **blockUQ;
  short **blockV, **blockVQ;

  short **YDCCoeffBuffer;
  short **UDCCoeffBuffer;
  short **VDCCoeffBuffer;

  int selectedQIdx, EGKLuma, EGKChroma;

  int compressionRateGroup; // 0: fixed compression rate for each block 1: fixed compression rate for a row of MBs
  double minCompressionRatio; // minimum compression ratio
  // Use block based compression for IBDI on
  // Use 12 bit to 8 bit compression for IBDI on
  if (img->bitdepth_luma >= 12)
  {
    compressionRateGroup = 0;
    minCompressionRatio = 1.5; 
  }
  else // use MB row compression for IBDI off
       // Use 8 bit to 4 bit compression for IBDI off
  {
    compressionRateGroup = 1;
    minCompressionRatio = 2; 
  }


  imageWidth = p->size_x;
  imageHeight = p->size_y;
  inputFrameY = p->imgY;
  inputFrameU = p->imgUV[0];
  inputFrameV = p->imgUV[1];
 
  numberOfBlockRows = imageHeight/blockHeight;
  numberOfBlockColumns = imageWidth/blockHeight;
  MBHeightInBlocks = (MBHeight/blockHeight);

  // Compute blockBitBudget for luma+chroma using blockHeight and minimumCompression ratio
  blockBitBudget = (int) ((blockHeight*blockHeight*(img->bitdepth_luma + img->bitdepth_chroma/2))/minCompressionRatio);
  
  numberOfTransformLevelsLuma = ilog2(blockHeight);
  numberOfTransformLevelsChroma = numberOfTransformLevelsLuma - 1;


  // allocate memory for blocks
  // Y
  blockY = (short **) malloc(blockHeight*sizeof(short *));
  blockYQ = (short **) malloc(blockHeight*sizeof(short *));
  for(i=0; i<blockHeight; i++)
  {
    blockY[i] = (short *) malloc(blockHeight*sizeof(short));
    blockYQ[i] = (short *) malloc(blockHeight*sizeof(short));
  }
  // U and V
  blockU = (short **) malloc((blockHeight/2)*sizeof(short *));
  blockUQ = (short **) malloc((blockHeight/2)*sizeof(short *));
  blockV = (short **) malloc((blockHeight/2)*sizeof(short *));
  blockVQ = (short **) malloc((blockHeight/2)*sizeof(short *));
  for(i=0; i<blockHeight/2; i++)
  {
    blockU[i] = (short *) malloc((blockHeight/2)*sizeof(short));
    blockUQ[i] = (short *) malloc((blockHeight/2)*sizeof(short));
    blockV[i] = (short *) malloc((blockHeight/2)*sizeof(short));
    blockVQ[i] = (short *) malloc((blockHeight/2)*sizeof(short));
  }

  // Allocate memory for DC coefficient buffers
  YDCCoeffBuffer = (short **) malloc(numberOfBlockRows*sizeof(short *));
  UDCCoeffBuffer = (short **) malloc(numberOfBlockRows*sizeof(short *));
  VDCCoeffBuffer = (short **) malloc(numberOfBlockRows*sizeof(short *));
  for(i=0; i<numberOfBlockRows; i++)
  {
    YDCCoeffBuffer[i] = (short *) malloc(numberOfBlockColumns*sizeof(short));
    UDCCoeffBuffer[i] = (short *) malloc(numberOfBlockColumns*sizeof(short));
    VDCCoeffBuffer[i] = (short *) malloc(numberOfBlockColumns*sizeof(short));
  }
  
  // loop over block rows
  for(i=0; i<numberOfBlockRows; i++)
  {
    // for the fixed MB row option
    if(compressionRateGroup)
    {
      if( i % MBHeightInBlocks == 0)
      {
        // Seek the bistream pointers
        compressedFrameBuf->bit_ptr = 8;
        compressedFrameBuf->buf_ptr = ( (blockBitBudget*i*numberOfBlockColumns)>>3 );     
      }
    }

    // loop over block columns
    for(j=0; j<numberOfBlockColumns; j++)
    {
      // Seek the bistream pointers for the fixed block size option
      if(!compressionRateGroup)
      {
        // Seek the bistream pointers
        compressedFrameBuf->bit_ptr = 8;
        compressedFrameBuf->buf_ptr = ( (blockBitBudget*(i*numberOfBlockColumns+j))>>3 );
      }

      // Entropy decoding
      ReadCompressedBlock(compressedFrameBuf, blockYQ, blockUQ, blockVQ, &selectedQIdx, &EGKLuma, &EGKChroma);

      if (compressionRateGroup)
      {
        // Reconstruct DC coefficients for Y,U,V
        ReconstructDCCoefficient(i, j, blockYQ, blockUQ, blockVQ, YDCCoeffBuffer, UDCCoeffBuffer, VDCCoeffBuffer);
        // Store the DC values to buffer 
        YDCCoeffBuffer[i][j] = blockYQ[0][0];
        UDCCoeffBuffer[i][j] = blockUQ[0][0];
        VDCCoeffBuffer[i][j] = blockVQ[0][0];
      }
      
      if (selectedQIdx > 0)
      {
        // Inverse quantize the transform coeff
        // Y
        InverseQuantization(selectedQIdx, blockHeight, blockY, blockYQ);
        // U
        InverseQuantization(selectedQIdx, blockHeight/2, blockU, blockUQ);
        // V
        InverseQuantization(selectedQIdx, blockHeight/2, blockV, blockVQ);

        // perform inverse 2-D DWT
        // Y
        UnrolledBackward2DTransform( blockHeight, numberOfTransformLevelsLuma, blockY);
        // U
        UnrolledBackward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockU);
        // V
        UnrolledBackward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockV);

        // store the block data to the image array and add 128
        StoreBlock(i, j, imageWidth, inputFrameY, inputFrameU, inputFrameV, blockY, blockU, blockV);
      }
      else
      {
        // perform inverse 2-D DWT
        // Y
        UnrolledBackward2DTransform( blockHeight, numberOfTransformLevelsLuma, blockYQ);
        // U
        UnrolledBackward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockUQ);
        // V
        UnrolledBackward2DTransform( blockHeight/2, numberOfTransformLevelsChroma, blockVQ);

        // store the block data to the image array and add 128
        StoreBlock(i, j, imageWidth, inputFrameY, inputFrameU, inputFrameV, blockYQ, blockUQ, blockVQ);
      }
    }
  }

  // Free memory for blocks
  // Y
  for(i=0; i<blockHeight; i++)
  {
    free(blockY[i]);
    free(blockYQ[i]);
  }
  free(blockY);
  free(blockYQ);

  // U and V
  for(i=0; i<blockHeight/2; i++)
  {
    free(blockU[i]);
    free(blockUQ[i]);
    free(blockV[i]);
    free(blockVQ[i]);
  }
  free(blockU);
  free(blockUQ);
  free(blockV);
  free(blockVQ);
  

  // free memory for DC coefficient buffers
  for(i=0; i<numberOfBlockRows; i++)
  {
    free(YDCCoeffBuffer[i]);
    free(UDCCoeffBuffer[i]);
    free(VDCCoeffBuffer[i]);
  }
  free(YDCCoeffBuffer);
  free(UDCCoeffBuffer);
  free(VDCCoeffBuffer);

  return;
}


/*!
************************************************************************
* \brief
*    Load Y, U, and V blocks from input image and subtract 128 
*
*  \par Input:
*     blockRowIdx: row index of the block
*     blockColumnIdx: column index of the block
*     imageWidth: image width in pixels
*     inputFrameY: pointer to input frame luma
*     inputFrameU: pointer to input frame chroma component U
*     inputFrameV: pointer to input frame chroma component V
*   \par Output:
*     blockY: pointer to luma block
*     blockU: pointer to chroma U block
*     blockV: pointer to chroma V block
*
************************************************************************
*/
void LoadBlock(int blockRowIdx, int blockColumnIdx, int imageWidth,
               unsigned short **inputFrameY, unsigned short **inputFrameU, unsigned short **inputFrameV,
               short **blockY, short **blockU, short **blockV)
{
  int i,j;
  int topLeftX, topLeftY;

  // compute coordinates of the top-left pixel of the block
  topLeftX = blockColumnIdx*blockHeight;
  topLeftY = blockRowIdx*blockHeight;

  // load Y block
  // loop over rows
  for(i=0; i<blockHeight; i++)
  {
    // loop  over columns
    for(j=0; j<blockHeight; j++)
    {
      blockY[i][j] = (short) inputFrameY[topLeftY+i][topLeftX+j];// - 128;
    }
  }

  // load U and V blocks
  // loop over rows
  for(i=0; i<blockHeight/2; i++)
  {
    // loop  over columns
    for(j=0; j<blockHeight/2; j++)
    {
      blockU[i][j] = (short) inputFrameU[topLeftY/2+i][topLeftX/2+j];// - 128;
      blockV[i][j] = (short) inputFrameV[topLeftY/2+i][topLeftX/2+j];// - 128;
    }
  }

  return;
}

/*!
************************************************************************
* \brief
*    Store Y, U, and V blocks to output image and add 128 and clip to [0, 255]
*
*  \par Input:
*     blockRowIdx: row index of the block
*     blockColumnIdx: column index of the block
*     imageWidth: image width in pixels
*     inputFrameY: pointer to input frame luma
*     inputFrameU: pointer to input frame chroma component U
*     inputFrameV: pointer to input frame chroma component V
*   \par Output:
*     blockY: pointer to luma block
*     blockU: pointer to chroma U block
*     blockV: pointer to chroma V block
*
************************************************************************
*/
void StoreBlock(int blockRowIdx, int blockColumnIdx, int imageWidth,
                unsigned short **inputFrameY, unsigned short **inputFrameU, unsigned short **inputFrameV,
                short **blockY, short **blockU, short **blockV)
{
  int i,j;
  int topLeftX, topLeftY;
  int maxVal, minVal;

  maxVal = (1<<(img->bitdepth_luma-1))-1;
  minVal = -(1<<(img->bitdepth_luma-1));

  // compute coordinates of the top-left pixel of the block
  topLeftX = blockColumnIdx*blockHeight;
  topLeftY = blockRowIdx*blockHeight;

  // load Y block
  // loop over rows
  for(i=0; i<blockHeight; i++)
  {
    // loop  over columns
    for(j=0; j<blockHeight; j++)
    {
      blockY[i][j] = (blockY[i][j]<minVal)?minVal:blockY[i][j];
      blockY[i][j] = (blockY[i][j]>maxVal)?maxVal:blockY[i][j];
      inputFrameY[topLeftY+i][topLeftX+j] =  (unsigned short) blockY[i][j] - minVal;
    }
  }

  // load U and V blocks
  // loop over rows
  for(i=0; i<blockHeight/2; i++)
  {
    // loop  over columns
    for(j=0; j<blockHeight/2; j++)
    {
      blockU[i][j] = (blockU[i][j]<minVal)?minVal:blockU[i][j];
      blockU[i][j] = (blockU[i][j]>maxVal)?maxVal:blockU[i][j];
      inputFrameU[topLeftY/2+i][topLeftX/2+j]  = (unsigned short) blockU[i][j] - minVal;

      blockV[i][j] = (blockV[i][j]<minVal)?minVal:blockV[i][j];
      blockV[i][j] = (blockV[i][j]>maxVal)?maxVal:blockV[i][j];
      inputFrameV[topLeftY/2+i][topLeftX/2+j]  = (unsigned short) blockV[i][j] - minVal;
    }
  }

  return;
}

/*!
************************************************************************
* \brief
*     Performs 2-D Lifting IWT
*
*  \par Input:
*     pBlockHeight: dimension of the input block in pixels
*     numberOfLevels: number of levels in the DWT
*   \par Input/Output:
*     block: pointer to block data
*
************************************************************************
*/
void UnrolledForward2DTransform(int pBlockHeight, int numberOfLevels, short **block)
{
  int i, j;

  ///////////////////////////////////////////////////
  // perform column-wise 1-D transform
  ///////////////////////////////////////////////////
  // loop over columns
  for(j=0; j<pBlockHeight; j++)
  {
    // Level 0
    block[1][j] = block[1][j] - block[0][j];
    block[0][j] = block[0][j] + block[1][j]/2;
    block[3][j] = block[3][j] - block[2][j];
    block[2][j] = block[2][j] + block[3][j]/2;
    if ( pBlockHeight == 8 )
    {
      block[5][j] = block[5][j] - block[4][j];
      block[4][j] = block[4][j] + block[5][j]/2;
      block[7][j] = block[7][j] - block[6][j];
      block[6][j] = block[6][j] + block[7][j]/2;
    }

    // Level 1
    block[2][j] = block[2][j] - block[0][j];
    block[0][j] = block[0][j] + block[2][j]/2;
    if ( pBlockHeight == 8 )
    {
      block[6][j] = block[6][j] - block[4][j];
      block[4][j] = block[4][j] + block[6][j]/2;

      // Level 2
      block[4][j] = block[4][j] - block[0][j];
      block[0][j] = block[0][j] + block[4][j]/2;
    }
  }

  ///////////////////////////////////////////////////
  // perform row-wise 1-D transform
  ///////////////////////////////////////////////////
  // loop over columns
  for(i=0; i<pBlockHeight; i++)
  {
    // Level 0
    block[i][1] = block[i][1] - block[i][0];
    block[i][0] = block[i][0] + block[i][1]/2;
    block[i][3] = block[i][3] - block[i][2];
    block[i][2] = block[i][2] + block[i][3]/2;
    if ( pBlockHeight == 8 )
    {
      block[i][5] = block[i][5] - block[i][4];
      block[i][4] = block[i][4] + block[i][5]/2;
      block[i][7] = block[i][7] - block[i][6];
      block[i][6] = block[i][6] + block[i][7]/2;
    }

    // Level 1
    block[i][2] = block[i][2] - block[i][0];
    block[i][0] = block[i][0] + block[i][2]/2;
    if ( pBlockHeight == 8 )
    {
      block[i][6] = block[i][6] - block[i][4];
      block[i][4] = block[i][4] + block[i][6]/2;

      // Level 2
      block[i][4] = block[i][4] - block[i][0];
      block[i][0] = block[i][0] + block[i][4]/2;
    }
  }
  
  block[0][0] = block[0][0] - (1<<(img->bitdepth_luma-1));
  return;
}

/*!
************************************************************************
* \brief
*     Performs 2-D Lifting inverse IWT
*
*  \par Input:
*     pBlockHeight: dimension of the input block in pixels
*     numberOfLevels: number of levels in the inverse DWT
*   \par Input/Output:
*     block: pointer to block data
*
************************************************************************
*/
void UnrolledBackward2DTransform(int pBlockHeight, int numberOfLevels, short **block)
{
  int i, j;
  
  ///////////////////////////////////////////////////
  // perform row-wise 1-D inverse transform
  ///////////////////////////////////////////////////
  // loop over rows
  for(i=0; i<pBlockHeight; i++)
  {
    if ( pBlockHeight == 8 )
    {
      // Level 2
      block[i][0] = block[i][0] - block[i][4]/2;
      block[i][4] = block[i][4] + block[i][0];
      
      // Level 1
      block[i][4] = block[i][4] - block[i][6]/2;
      block[i][6] = block[i][6] + block[i][4];
    }

    // Level 1
    block[i][0] = block[i][0] - block[i][2]/2;
    block[i][2] = block[i][2] + block[i][0];
    
    // Level 0
    block[i][0] = block[i][0] - block[i][1]/2;
    block[i][1] = block[i][1] + block[i][0];
    block[i][2] = block[i][2] - block[i][3]/2;
    block[i][3] = block[i][3] + block[i][2];
    if ( pBlockHeight == 8 )
    {
      block[i][4] = block[i][4] - block[i][5]/2;
      block[i][5] = block[i][5] + block[i][4];
      block[i][6] = block[i][6] - block[i][7]/2;
      block[i][7] = block[i][7] + block[i][6];
    }
  }
  
  ///////////////////////////////////////////////////
  // perform column-wise 1-D inverse transform
  ///////////////////////////////////////////////////
  // loop over columns
  for(j=0; j<pBlockHeight; j++)
  {
    if ( pBlockHeight == 8 )
    {
      // Level 2
      block[0][j] = block[0][j] - block[4][j]/2;
      block[4][j] = block[4][j] + block[0][j];

      // Level 1
      block[4][j] = block[4][j] - block[6][j]/2;
      block[6][j] = block[6][j] + block[4][j];
    }

    // Level 1
    block[0][j] = block[0][j] - block[2][j]/2;
    block[2][j] = block[2][j] + block[0][j];

    // Level 0
    block[0][j] = block[0][j] - block[1][j]/2;
    block[1][j] = block[1][j] + block[0][j];
    block[2][j] = block[2][j] - block[3][j]/2;
    block[3][j] = block[3][j] + block[2][j];
    if ( pBlockHeight == 8 )
    {
      block[4][j] = block[4][j] - block[5][j]/2;
      block[5][j] = block[5][j] + block[4][j];
      block[6][j] = block[6][j] - block[7][j]/2;
      block[7][j] = block[7][j] + block[6][j];
    }
  }
  return;
}

/*!
************************************************************************
* \brief
*     Quantizes the transform coefficients except the lowest frequency
*
*  \par Input:
*     QIdx: quantization index (QLevel = 2^QIdx)
*     pBlockHeight: dimension of the input block
*     block: pointer to input block data
*   \par Output:
*     blockQ: pointer to output quantized block data
*
************************************************************************
*/
void Quantization(int QIdx, int pBlockHeight, short** block, short** blockQ)
{
  int i, j, k, l;
  int log2QLevel, QLevel, freqLevel;
  int QIdx2log2QLevel[numberOfQuantizationIndices][4] = 
  {{0,  0,  0,  0},
   {0,  0,  0,  1},
   {0,  0,  1,  2},
   {0,  1,  2,  3},
   {1,  2,  3,  4},
   {2,  3,  4,  5},
   {3,  4,  5,  6},
   {4,  5,  6,  7},
   {5,  6,  7,  8},
   {6,  7,  8,  9},
   {7,  8,  9,  10},
   {8,  9,  10, 11},
   {9,  10, 11, 12},
   {10, 11, 12, 12},
   {11, 12, 12, 12},
   {12, 12, 12, 12}};
  // loop over rows
  for(i=0; i<pBlockHeight; i++)
  {
    // loop over columns
    for(j=0; j<pBlockHeight; j++)
    {
      if(i+j>0)
      {
        // Determine frequency level
        // L   H4  H2  H4  H0  H4  H2  H4
        // H4  H5  H4  H5  H4  H5  H4  H5
        // H2  H4  H3  H4  H2  H4  H2  H4
        // H4  H5  H4  H5  H4  H5  H4  H5
        // H0  H4  H2  H4  H1  H4  H2  H4
        // H4  H5  H4  H5  H4  H5  H4  H5
        // H2  H4  H3  H4  H2  H4  H3  H4
        // H4  H5  H4  H5  H4  H5  H4  H5

        for(k=1; k<3; k++)
        {
          l = (1<<(k+1));
          if(i%l > 0 && j%l > 0)
          {
            freqLevel = (k-1)*2;
            break;
          }
          else if(i%l > 0 || j%l > 0)
          {
            freqLevel = (k-1)*2+1;
            break;
          }
        }

        freqLevel = 3-freqLevel;

        log2QLevel = QIdx2log2QLevel[QIdx][freqLevel];
        QLevel = (1<<log2QLevel);

        // Mid-tread quantization 
        if(block[i][j] >= 0)
        {
          blockQ[i][j] = ( ( (block[i][j]<<1) + QLevel ) >> (log2QLevel+1) );
        }
        else
        {
          blockQ[i][j] = -( ( (abs(block[i][j])<<1) + QLevel ) >> (log2QLevel+1) );
        }
      }
      else
      {
        blockQ[i][j] = block[i][j];
      }
    }
  }
  return;
}

/*!
************************************************************************
* \brief
*     Inverse quantizes the transform coefficients except the lowest frequency
*
*  \par Input:
*     QIdx: quantization index (QLevel = 2^QIdx)
*     pBlockHeight: dimension of the input block
*     blockQ: pointer to input quantized block data
*   \par Output:
*     block: pointer to output block data
*
************************************************************************
*/
void InverseQuantization(int QIdx, int pBlockHeight, short** block, short** blockQ)
{
  int i, j, k, l;
  int log2QLevel, QLevel, freqLevel;
  int QIdx2log2QLevel[numberOfQuantizationIndices][4] = 
  {{0,  0,  0,  0},
   {0,  0,  0,  1},
   {0,  0,  1,  2},
   {0,  1,  2,  3},
   {1,  2,  3,  4},
   {2,  3,  4,  5},
   {3,  4,  5,  6},
   {4,  5,  6,  7},
   {5,  6,  7,  8},
   {6,  7,  8,  9},
   {7,  8,  9,  10},
   {8,  9,  10, 11},
   {9,  10, 11, 12},
   {10, 11, 12, 12},
   {11, 12, 12, 12},
   {12, 12, 12, 12}};

  // loop over rows
  for(i=0; i<pBlockHeight; i++)
  {
    // loop over columns
    for(j=0; j<pBlockHeight; j++)
    {
      if(i+j>0)
      {
        // Determine frequency level
        // L   H4  H2  H4  H0  H4  H2  H4
        // H4  H5  H4  H5  H4  H5  H4  H5
        // H2  H4  H3  H4  H2  H4  H2  H4
        // H4  H5  H4  H5  H4  H5  H4  H5
        // H0  H4  H2  H4  H1  H4  H2  H4
        // H4  H5  H4  H5  H4  H5  H4  H5
        // H2  H4  H3  H4  H2  H4  H3  H4
        // H4  H5  H4  H5  H4  H5  H4  H5

        for(k=1; k<3; k++)
        {
          l = (1<<(k+1));
          if(i%l > 0 && j%l > 0)
          {
            freqLevel = (k-1)*2;
            break;
          }
          else if(i%l > 0 || j%l > 0)
          {
            freqLevel = (k-1)*2+1;
            break;
          }
        }

        freqLevel = 3-freqLevel;

        log2QLevel = QIdx2log2QLevel[QIdx][freqLevel];
        QLevel = (1<<log2QLevel);

        block[i][j] = QLevel*blockQ[i][j];
      }
      else
      {
        block[i][j] = blockQ[i][j];
      }
    }
  }
  return;
}

/*!
************************************************************************
* \brief
*     Computes the entropy coded bit size of Y coefficients
*     Exp-Golomb option (EG0 or EG3) with best compression is selected
*
*  \par Input:
*     blockYQ: pointer to input quantized block data
*  \par Input/Output:
*     EGKLuma: selected ExpGolomb coding k parameter
*  \par Output:
*     entropy coded bit size
*
************************************************************************
*/
int ComputeEntropyCodedBitSizeY(short **blockYQ, int *EGKLuma)
{
  int i, j;
  int EG0Size, EG3Size;
  unsigned int mappedCoeffValue;
  EG0Size = 0;
  EG3Size = 0;

  // loop over rows
  for(i=0; i<blockHeight; i++)
  {
    // loop over columns
    for(j=0; j<blockHeight; j++)
    {
      // map quantized coefficient levels to non-negative integers
      if( blockYQ[i][j] >= 0 )
      {
        mappedCoeffValue = (blockYQ[i][j] << 1);
      }
      else
      {
        mappedCoeffValue = (abs(blockYQ[i][j]) << 1) - 1;
      }
      // Increment the bit-counter for EG0 coding
      EG0Size += 2*( ilog2( mappedCoeffValue + 1 ) ) + 1;
      // Increment the bit-counter for EG3 coding
      EG3Size += 2*( ilog2( mappedCoeffValue/8 + 1 ) ) + 4;  
    }
  }

  if(EG3Size < EG0Size)
  {
    *EGKLuma = 3;
    return EG3Size;
  }
  else
  {
    *EGKLuma = 0;
    return EG0Size;
  }
}

/*!
************************************************************************
* \brief
*     Computes the entropy coded bit size of U and V coefficients
*     Exp-Golomb option (EG0 or EG3) with best compression is selected
*
*  \par Input:
*     blockUQ: pointer to input quantized U block data
*     blockVQ: pointer to input quantized V block data
*  \par Input/Output:
*     EGKLuma: selected ExpGolomb coding k parameter
*  \par Output:
*     entropy coded bit size
*
************************************************************************
*/
int ComputeEntropyCodedBitSizeUV(short **blockUQ, short **blockVQ, int *EGKChroma)
{
  int i, j;
  int EG0Size, EG3Size;
  unsigned int mappedCoeffValue;
  EG0Size = 0;
  EG3Size = 0;

  // loop over rows
  for(i=0; i<blockHeight/2; i++)
  {
    // loop over columns
    for(j=0; j<blockHeight/2; j++)
    {
      // U
      // map quantized coefficient levels to non-negative integers
      if( blockUQ[i][j] >= 0 )
      {
        mappedCoeffValue = (blockUQ[i][j] << 1);
      }
      else
      {
        mappedCoeffValue = (abs(blockUQ[i][j]) << 1) - 1;
      }
      // Increment the bit-counter for EG0 coding
      EG0Size += 2*( ilog2( mappedCoeffValue + 1 ) ) + 1;
      // Increment the bit-counter for EG3 coding
      EG3Size += 2*( ilog2( mappedCoeffValue/8 + 1 ) ) + 4;


      // V
      // map quantized coefficient levels to non-negative integers
      if( blockVQ[i][j] >= 0 )
      {
        mappedCoeffValue = (blockVQ[i][j] << 1);
      }
      else
      {
        mappedCoeffValue = (abs(blockVQ[i][j]) << 1) - 1;
      }
      // Increment the bit-counter for EG0 coding
      EG0Size += 2*( ilog2( mappedCoeffValue + 1 ) ) + 1;
      // Increment the bit-counter for EG3 coding
      EG3Size += 2*( ilog2( mappedCoeffValue/8 + 1 ) ) + 4;
    }
  }

  if(EG3Size < EG0Size)
  {
    *EGKChroma = 3;
    return EG3Size;
  }
  else
  {
    *EGKChroma = 0;
    return EG0Size;
  }
}

/*!
************************************************************************
* \brief
*     Computes DC coefficient residual for Y,U,V
*
*  \par Input:
*     blockRowIdx: block row index
*     blockColumnIdx: block column index
*     YDCCoeffBuffer: pointer to Y DC coefficient buffer
*     UDCCoeffBuffer: pointer to U DC coefficient buffer
*     VDCCoeffBuffer: pointer to V DC coefficient buffer
*  \par Input/Output
*     blockY: pointer to Y block
*     blockU: pointer to U block
*     blockV: pointer to V block
*
************************************************************************
*/
void ComputeDCCoefficientResidual(int blockRowIdx, int blockColumnIdx, short **blockY, short **blockU, short **blockV, short **YDCCoeffBuffer, short **UDCCoeffBuffer, short **VDCCoeffBuffer)
{
  int leftAvailable, upperAvailable;
  int predictedDCY, predictedDCU, predictedDCV;
  int thread1, thread2, thread3;
  int MBHeightInBlocks;
  MBHeightInBlocks = (MBHeight/blockHeight);

  // Determine availability
  if(blockColumnIdx > 0)
  {
    leftAvailable = 1;
  }
  else
  {
    leftAvailable = 0;
  }
  if( (blockRowIdx%MBHeightInBlocks) > 0) // do not use upper blocks for predicting top row of MBs
  {
    upperAvailable = 1;
  }
  else
  {
    upperAvailable = 0;
  }

  // A B D
  // C X
  // thread(X) = min{|A-B|, |A-C|, |A-(B+C)/2|}
  // pred(X) = C, if thread(X) = |A-B|
  //         = B, if thread(X) = |A-C|
  //         = (B+C)/2, if thread(X) = |A-(B+C)/2|
  //         = C, if B is not available
  //         = B, if C is not available
  //         = 0, if B and C are not available
  
  // Compute pred(X)
  if(leftAvailable && upperAvailable)
  {
    // Y
    thread1 = abs(YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] );
    thread2 = abs(YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1] );
    thread3 = abs(YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - ((YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1) );
    if(thread1 < thread2 && thread1 < thread3)
    {
      predictedDCY = YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    }
    else if (thread2 < thread3)
    {
      predictedDCY = YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    }
    else
    {
      predictedDCY = ((YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1);
    }

    // U
    thread1 = abs(UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] );
    thread2 = abs(UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1] );
    thread3 = abs(UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - ((UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1) );
    if(thread1 < thread2 && thread1 < thread3)
    {
      predictedDCU = UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    }
    else if (thread2 < thread3)
    {
      predictedDCU = UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    }
    else
    {
      predictedDCU = ((UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1);
    }

    // V
    thread1 = abs(VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] );
    thread2 = abs(VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1] );
    thread3 = abs(VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - ((VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1) );
    if(thread1 < thread2 && thread1 < thread3)
    {
      predictedDCV = VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    }
    else if (thread2 < thread3)
    {
      predictedDCV = VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    }
    else
    {
      predictedDCV = ((VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1);
    }

  }
  else if (leftAvailable)
  {
    predictedDCY = YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    predictedDCU = UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    predictedDCV = VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
  }
  else if (upperAvailable)
  {
    predictedDCY = YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    predictedDCU = UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    predictedDCV = VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
  }
  else
  {
    predictedDCY = 0;
    predictedDCU = 0;
    predictedDCV = 0;
  }

  
  // Compute residual(X) = X - pred(X)
  blockY[0][0] -= predictedDCY;
  blockU[0][0] -= predictedDCU;
  blockV[0][0] -= predictedDCV;
  
  return;
}

/*!
************************************************************************
* \brief
*     Reconstructs DC coefficient for Y,U,V from residual and prediction
*
*  \par Input:
*     blockRowIdx: block row index
*     blockColumnIdx: block column index
*     YDCCoeffBuffer: pointer to Y DC coefficient buffer
*     UDCCoeffBuffer: pointer to U DC coefficient buffer
*     VDCCoeffBuffer: pointer to V DC coefficient buffer
*  \par Input/Output
*     blockY: pointer to Y block
*     blockU: pointer to U block
*     blockV: pointer to V block
*
************************************************************************
*/
void ReconstructDCCoefficient(int blockRowIdx, int blockColumnIdx, short **blockY, short **blockU, short **blockV, short **YDCCoeffBuffer, short **UDCCoeffBuffer, short **VDCCoeffBuffer)
{
  int leftAvailable, upperAvailable;
  int predictedDCY, predictedDCU, predictedDCV;
  int thread1, thread2, thread3;
  int MBHeightInBlocks;
  MBHeightInBlocks = (MBHeight/blockHeight);

  // Determine availability
  if(blockColumnIdx > 0)
  {
    leftAvailable = 1;
  }
  else
  {
    leftAvailable = 0;
  }
  if( (blockRowIdx%MBHeightInBlocks) > 0) // do not use upper blocks for predicting top row of MBs
  {
    upperAvailable = 1;
  }
  else
  {
    upperAvailable = 0;
  }

  // A B D
  // C X
  // thread(X) = min{|A-B|, |A-C|, |A-(B+C)/2|}
  // pred(X) = C, if thread(X) = |A-B|
  //         = B, if thread(X) = |A-C|
  //         = (B+C)/2, if thread(X) = |A-(B+C)/2|
  //         = C, if B is not available
  //         = B, if C is not available
  //         = 0, if B and C are not available
  
  // Compute pred(X)
  if(leftAvailable && upperAvailable)
  {
    // Y
    thread1 = abs(YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] );
    thread2 = abs(YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1] );
    thread3 = abs(YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - ((YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1) );
    if(thread1 < thread2 && thread1 < thread3)
    {
      predictedDCY = YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    }
    else if (thread2 < thread3)
    {
      predictedDCY = YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    }
    else
    {
      predictedDCY = ((YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1);
    }

    // U
    thread1 = abs(UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] );
    thread2 = abs(UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1] );
    thread3 = abs(UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - ((UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1) );
    if(thread1 < thread2 && thread1 < thread3)
    {
      predictedDCU = UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    }
    else if (thread2 < thread3)
    {
      predictedDCU = UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    }
    else
    {
      predictedDCU = ((UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1);
    }

    // V
    thread1 = abs(VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] );
    thread2 = abs(VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1] );
    thread3 = abs(VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx-1] - ((VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1) );
    if(thread1 < thread2 && thread1 < thread3)
    {
      predictedDCV = VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    }
    else if (thread2 < thread3)
    {
      predictedDCV = VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    }
    else
    {
      predictedDCV = ((VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx] + VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1])>>1);
    }

  }
  else if (leftAvailable)
  {
    predictedDCY = YDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    predictedDCU = UDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
    predictedDCV = VDCCoeffBuffer[blockRowIdx][blockColumnIdx-1];
  }
  else if (upperAvailable)
  {
    predictedDCY = YDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    predictedDCU = UDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
    predictedDCV = VDCCoeffBuffer[blockRowIdx-1][blockColumnIdx];
  }
  else
  {
    predictedDCY = 0;
    predictedDCU = 0;
    predictedDCV = 0;
  }
  
  // Reconstruct X = residual(X) + pred(X)
  blockY[0][0] += predictedDCY;
  blockU[0][0] += predictedDCU;
  blockV[0][0] += predictedDCV;
  
  return;
}

/*!
************************************************************************
* \brief
*     Entropy encodes and writes compressed block to the buffer
* 
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*     blockYQ: pointer to input quantized Y block data
*     blockUQ: pointer to input quantized U block data
*     blockVQ: pointer to input quantized V block data
*     selectedQIdx: pointer to selected quantization index
*     EGKLuma: pointer to selected ExpGolomb parameter (EGk) for luma
*     EGKChroma: pointer to selected ExpGolomb parameter (EGk) for chroma
*
************************************************************************
*/
void WriteCompressedBlock(CFB_Bitstream *stream, short **blockYQ, short **blockUQ, short **blockVQ, int *selectedQIdx, int *EGKLuma, int *EGKChroma)
{
  int i, j;
  unsigned int mappedCoeffValue;
  
  // write quantization index using unary coding
  PutBits(stream, (*selectedQIdx + 1), 1);
  // write EGK values with fixed length coding
  if(*EGKLuma == 0)
  {
    PutBits(stream, 1, 0);
  }
  else
  {
    PutBits(stream, 1, 1);
  }
  if(*EGKChroma == 0)
  {
    PutBits(stream, 1, 0);
  }
  else
  {
    PutBits(stream, 1, 1);
  }
   
  // write quantized luma transform coefficients
  for(i=0; i<blockHeight; i++)
  {
    for(j=0; j<blockHeight; j++)
    {
      // map quantized coefficient levels to non-negative integers
      if( blockYQ[i][j] >= 0 )
      {
        mappedCoeffValue = (blockYQ[i][j] << 1);
      }
      else
      {
        mappedCoeffValue = (abs(blockYQ[i][j]) << 1) - 1;
      }
      UnsignedExpGolombEncoding(stream, *EGKLuma, mappedCoeffValue);
    }
  }
  
  // write quantized chroma U transform coefficients
  for(i=0; i<blockHeight/2; i++)
  {
    for(j=0; j<blockHeight/2; j++)
    {
      // map quantized coefficient levels to non-negative integers
      if( blockUQ[i][j] >= 0 )
      {
        mappedCoeffValue = (blockUQ[i][j] << 1);
      }
      else
      {
        mappedCoeffValue = (abs(blockUQ[i][j]) << 1) - 1;
      }
      UnsignedExpGolombEncoding(stream, *EGKChroma, mappedCoeffValue);
    }
  }

  // write quantized chroma V transform coefficients
  for(i=0; i<blockHeight/2; i++)
  {
    for(j=0; j<blockHeight/2; j++)
    {
      // map quantized coefficient levels to non-negative integers
      if( blockVQ[i][j] >= 0 )
      {
        mappedCoeffValue = (blockVQ[i][j] << 1);
      }
      else
      {
        mappedCoeffValue = (abs(blockVQ[i][j]) << 1) - 1;
      }
      UnsignedExpGolombEncoding(stream, *EGKChroma, mappedCoeffValue);
    }
  }
  
  return;
}

/*!
************************************************************************
* \brief
*     Entropy encodes and writes compressed block to the buffer
* 
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*  \par Output:
*     blockYQ: pointer to output quantized Y block data
*     blockUQ: pointer to output quantized U block data
*     blockVQ: pointer to output quantized V block data
*     selectedQIdx: pointer to selected quantization index
*     EGKLuma: pointer to selected ExpGolomb parameter (EGk) for luma
*     EGKChroma: pointer to selected ExpGolomb parameter (EGk) for chroma
*
************************************************************************
*/
void ReadCompressedBlock(CFB_Bitstream *stream, short **blockYQ, short **blockUQ, short **blockVQ, int *selectedQIdx, int *EGKLuma, int *EGKChroma)
{
  int i, j;
  unsigned int mappedCoeffValue;
  int N, bits;
  unsigned int temp;
  // read quantization index using unary coding
  temp = NextBits(stream, 16);
  bits = 15;
  for (N = 1; ((temp >> bits) & 0x1) != 1; N++, bits--);
  FlushBits(stream, N);
  *selectedQIdx = N-1;

  // read EGK values with fixed length coding
  *EGKLuma = ReadBits(stream, 1);
  if(*EGKLuma == 1)
  {
    *EGKLuma = 3;
  }
  *EGKChroma = ReadBits(stream, 1);
  if(*EGKChroma == 1)
  {
    *EGKChroma = 3;
  }


  // read quantized luma transform coefficients
  for(i=0; i<blockHeight; i++)
  {
    for(j=0; j<blockHeight; j++)
    {
      mappedCoeffValue = UnsignedExpGolombDecoding(stream, *EGKLuma); 
      // map quantized coefficient levels to non-negative integers
      if( mappedCoeffValue %2 == 0 )
      {
        blockYQ[i][j] = ( mappedCoeffValue >> 1);
      }
      else
      {
        blockYQ[i][j] = -((int)( (mappedCoeffValue + 1) >> 1));
      }
    }
  }

  // read quantized chroma U transform coefficients
  for(i=0; i<blockHeight/2; i++)
  {
    for(j=0; j<blockHeight/2; j++)
    {
      mappedCoeffValue = UnsignedExpGolombDecoding(stream, *EGKChroma); 
      // map quantized coefficient levels to non-negative integers
      if( mappedCoeffValue %2 == 0 )
      {
        blockUQ[i][j] = ( mappedCoeffValue >> 1);
      }
      else
      {
        blockUQ[i][j] = -((int)( (mappedCoeffValue + 1) >> 1));
      }
    }
  }

  // read quantized chroma V transform coefficients
  for(i=0; i<blockHeight/2; i++)
  {
    for(j=0; j<blockHeight/2; j++)
    {
      mappedCoeffValue = UnsignedExpGolombDecoding(stream, *EGKChroma); 
      // map quantized coefficient levels to non-negative integers
      if( mappedCoeffValue %2 == 0 )
      {
        blockVQ[i][j] = ( mappedCoeffValue >> 1);
      }
      else
      {
        blockVQ[i][j] = -((int)( (mappedCoeffValue + 1) >> 1));
      }
    }
  }

  return;
}



/*!
************************************************************************
* \brief
*     Unsigned ExpGolomb encoding
* 
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*     k: ExpGolomb parameter (EGk)
*     data: value of the data to encode
*
************************************************************************
*/
void UnsignedExpGolombEncoding(CFB_Bitstream * stream, int k, unsigned int data)
{

  int N;
  unsigned int CodeNum;
  unsigned int buf;
  unsigned int INFO;

  CodeNum = data + (1 << k);
  buf = CodeNum >> k;
  for (N = 0; buf > 0; N++, buf >>= 1) ;

  INFO = (unsigned int) (CodeNum - (1 << (N - 1 + k)));

  PutBits(stream, N, 1);
  if (N - 1 + k)
    PutBits(stream, (int) (N - 1 + k), INFO);
}


/*!
************************************************************************
* \brief
*     Unsigned ExpGolomb decoding
*
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*     k: ExpGolomb parameter (EGk)
*  \par Output:
*     data: value of the decoded data
*
************************************************************************
*/
unsigned int UnsignedExpGolombDecoding(CFB_Bitstream * stream, int k)
{

  //  CodeNum = 2^(N-1+k)-2^k+ INFO (read_bits_avsp2(N-1+k))
  int N, bits;
  unsigned int CodeNum;
  unsigned int INFO, temp;

  /* ==================================================================== */
  /* compute code length                         */
  /* ==================================================================== */
  temp = NextBits(stream, 16);
  
  bits = 15;
  for (N = 1; ((temp >> bits) & 0x1) != 1; N++, bits--);
  FlushBits(stream, N);
  
  if (N - 1 + k)
    INFO = ReadBits(stream, (int) (N - 1 + k));
  else
    INFO = 0;

  CodeNum = (1 << (N - 1 + k)) - (1 << k) + INFO;

  return CodeNum;
}

/*!
************************************************************************
* \brief
*     Put bits
*
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*     nrbits: number of bits to put
*     code: code word to put
*
************************************************************************
*/
void PutBits(CFB_Bitstream * stream, int nrbits, unsigned int code)
{

  unsigned int buffer;
  int sign, bit_ptr, buf_ptr;
  
  bit_ptr = stream->bit_ptr;
  buf_ptr = stream->buf_ptr;
  buffer = stream->databuf[buf_ptr];


  bit_ptr -= nrbits;
  sign = (bit_ptr > 0);
  if (sign)
    buffer ^= (code << bit_ptr);
  if (!sign)
    buffer ^= (code >> (-bit_ptr));

  stream->databuf[buf_ptr] = (unsigned char) buffer;

  while (bit_ptr <= 0) {
    buf_ptr++;
    bit_ptr += 8;
    if (bit_ptr > 0)
      stream->databuf[buf_ptr] = (code << bit_ptr) & 0xff;
    else
      stream->databuf[buf_ptr] = (code >> (-bit_ptr)) & 0xff;
  }


  stream->bit_ptr = bit_ptr;
  stream->buf_ptr = buf_ptr;

}

/*!
************************************************************************
* \brief
*     Read bits
*
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*     nrbits: number of bits to read
*  \par Output:
*     code: read code
*
************************************************************************
*/
unsigned int ReadBits(CFB_Bitstream * stream, int nrbits)
{

  unsigned int pattern;
  unsigned int buffer;
  int bit_ptr, buf_ptr;


  bit_ptr = stream->bit_ptr;
  buf_ptr = stream->buf_ptr;
  buffer = stream->databuf[buf_ptr];

  bit_ptr = (int) (bit_ptr - nrbits);

  while (bit_ptr <= 0) {
    bit_ptr += 8;
    buffer = buffer << 8;
    stream->buf_ptr = (stream->buf_ptr == stream->bufSize - 1)?(stream->buf_ptr - 1):(stream->buf_ptr);
    buffer += stream->databuf[(++buf_ptr)];
  }

  pattern = (unsigned int) ((buffer >> bit_ptr) & bhMask[nrbits]);

  stream->bit_ptr = bit_ptr;
  stream->buf_ptr = (int) (buf_ptr);
  stream->buf_ptr = (stream->buf_ptr == stream->bufSize)?(stream->buf_ptr - 1):(stream->buf_ptr);
  return pattern;
}

/*!
************************************************************************
* \brief
*     Peek next nrbits bits
*
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*     nrbits: number of bits to peek
*  \par Output:
*     code: read code
*
************************************************************************
*/
unsigned int NextBits(CFB_Bitstream * stream, int nrbits)
{

  unsigned int pattern;
  unsigned int buffer;
  int bit_ptr, buf_ptr;

  bit_ptr = stream->bit_ptr;
  buf_ptr = stream->buf_ptr;
  buffer = stream->databuf[buf_ptr];

  bit_ptr = (int) (bit_ptr - nrbits);

  while (bit_ptr <= 0) {
    bit_ptr += 8;
    buffer = buffer << 8;
    stream->buf_ptr = (stream->buf_ptr == stream->bufSize - 1)?(stream->buf_ptr - 1):(stream->buf_ptr);
    buffer += stream->databuf[(++buf_ptr)];
  }

  pattern = (unsigned int) ((buffer >> bit_ptr) & bhMask[nrbits]);

  return pattern;
}

/*!
************************************************************************
* \brief
*     Jump next nrbits bits
*
*  \par Input:
*     stream: pointer to compressed frame buffer structure
*     nrbits: number of bits jump
*
************************************************************************
*/
void FlushBits(CFB_Bitstream * stream, int nrbits)
{
  stream->bit_ptr = (int) (stream->bit_ptr - nrbits);
  while (stream->bit_ptr <= 0) {
    stream->bit_ptr += 8;
    stream->buf_ptr = (stream->buf_ptr + 1);
    stream->buf_ptr = (stream->buf_ptr == stream->bufSize)?(stream->buf_ptr - 1):(stream->buf_ptr);
  }
}

/*!
************************************************************************
* \brief
*     computes integer log2 of unsigned integers in the [0, 32767] range
*
*  \par Input:
*     x
*  \par Output:
*     log2(x)
*
************************************************************************
*/
unsigned int ilog2(register unsigned int x) 
{
  register unsigned int l=0;
  if(x >= 1<<8) { x>>=8; l|=8; }
  if(x >= 1<<4) { x>>=4; l|=4; }
  if(x >= 1<<2) { x>>=2; l|=2; }
  if(x >= 1<<1) l|=1;
  return l;
}

